-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore ALU.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Arithmetic-Logic-Unit for pre-incrementing absolute data memory access.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY alu IS
PORT (reset     : IN    STD_LOGIC; 
      clk       : IN    STD_LOGIC;
      clk_en    : IN    STD_LOGIC;
      inst      : IN    inst_bus;
      condition : IN    STD_LOGIC;
      token     : IN    STD_LOGIC;
      uBus      : IN    data_port;
      tos       : IN    data_bus;
      nos       : IN    data_bus;
      stack     : IN    data_bus;
      pc        : IN    data_bus;
      rsp       : IN    data_bus;
      task      : IN    data_bus;
      stBus     : INOUT status_rec;
      tor_zero  : OUT   STD_LOGIC;
      alu       : OUT   data_bus;
      tor       : OUT   data_bus);
END alu;

ARCHITECTURE rtl OF alu IS

SIGNAL i_group    : inst_group;
SIGNAL i_stack    : inst_stack;
SIGNAL i_type     : inst_type;
SIGNAL i_lit      : STD_LOGIC;
SIGNAL s_lit      : STD_LOGIC;

SIGNAL uBus_data  : data_bus;
SIGNAL adder_in1  : STD_LOGIC_VECTOR(data_width   DOWNTO 0); -- extended for DIVS
SIGNAL adder_mux  : STD_LOGIC_VECTOR(data_width   DOWNTO 0); -- extended for DIVS
SIGNAL result     : STD_LOGIC_VECTOR(data_width+1 DOWNTO 0); -- extended for DIVS step
SIGNAL sum        : data_bus;
SIGNAL alu_i      : data_bus;
SIGNAL carry_in   : STD_LOGIC;
SIGNAL carry_out  : STD_LOGIC;
SIGNAL carry      : STD_LOGIC;
SIGNAL sign       : STD_LOGIC;
SIGNAL step_carry : STD_LOGIC;
SIGNAL s_c        : STD_LOGIC;
SIGNAL tor_reg    : data_bus;
SIGNAL tor_new    : data_bus;
SIGNAL overflow   : STD_LOGIC;
SIGNAL ovl_new    : STD_LOGIC;
SIGNAL s_ovl      : STD_LOGIC;

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);
s_lit   <= stBus.flags(s_lit_bit);
uBus_data <= uBus.data;

alu <= alu_i;
tor <= tor_reg;
tor_zero <= '1' WHEN  with_tor='1' AND tor_reg=0  ELSE  '0';

-- adder_in1 is tos and is prepared by the process alu_output
-- extended adder for the DIVS instruction
result     <= adder(adder_in1, adder_mux, carry_in);
sum        <= result(data_width-1 DOWNTO 0);
sign       <= result(data_width-1);
step_carry <= result(data_width);
carry      <= result(data_width+1);
overflow   <= step_carry XOR sign;

-----------------------------------------------------------------------
-- alu_output: 
-----------------------------------------------------------------------

alu_output: PROCESS(i_lit, s_lit, i_type, i_stack, i_group, pc, tos, nos, sum, s_c, rsp, tor_reg, task, uBus_data, condition, carry, overflow, s_ovl, stack, token, step_carry)
BEGIN

adder_in1 <= tos(tos'high) & tos; -- sign extend TOS
adder_mux <= '0' & pc;
alu_i     <= sum;
carry_in  <= '0';
carry_out <= s_c;
tor_new   <= tor_reg;
ovl_new   <= s_ovl;

IF  i_lit='0'  THEN
   CASE  i_type  IS
   WHEN op_BRA =>
       CASE i_stack IS
       WHEN op_NONE => IF  with_tor='1' AND condition='1'  THEN
                          tor_new <= uBus_data; -- pop_rstack;
                       END IF;
       WHEN op_POP  => IF  with_tor='1' AND i_group=op_NZERO  THEN  -- NEXT
                          IF  tor_reg=0  THEN
                             tor_new <= uBus_data; --  pop_rstack;
                          ELSE
                             tor_new <= tor_reg-1;
                          END IF;
                       END IF;
       WHEN op_PUSH =>
          CASE i_group IS
          WHEN op_IRET  => IF  with_tor='1'  THEN
                              tor_new <= uBus_data;
                           END IF;
          WHEN op_INT | op_EXC =>
                           IF  with_tor='1'  THEN
                              tor_new <= pc;
                           END IF;
          WHEN op_QOVL  => IF  with_tor='1' AND s_ovl='1'  THEN
                              tor_new <= pc;
                           END IF;
          WHEN op_TOKEN => IF  with_ip='1' AND with_tor='1' AND with_tokens='1'  THEN
                              IF  token='0'  THEN
                                 tor_new <= tos; -- push_rstack;
                              END IF;
                           END IF;
          WHEN OTHERS => NULL;
          END CASE;
       WHEN OTHERS  => -- op_BOTH
                       IF  with_tor='1' AND condition='1'  THEN
                          tor_new <= pc; -- push_rstack;
                       END IF;
                       IF  s_lit='0'  THEN
                          alu_i <= tos;
                       END IF;
       END CASE;
   WHEN op_ALU =>
       CASE i_stack IS
       WHEN op_BOTH =>
           CASE  i_group  IS
           WHEN op_NOT  => alu_i <= NOT tos;
           WHEN op_SL   => alu_i <= tos(data_width-2 DOWNTO 0) & '0';  -- logical shift left: 0 goes into 0th bit pos.
                           carry_out <= tos(data_width-1);             -- this can be used for asl too.
           WHEN op_ASR  => alu_i <= tos(data_width-1) & tos(data_width-1 DOWNTO 1); -- preserve the sign.
                           carry_out <= tos(0);
           WHEN op_LSR  => alu_i <= '0' & tos(data_width-1 DOWNTO 1);  -- LSR: 0 goes into MSB position.
                           carry_out <= tos(0);
           WHEN op_ROR  => alu_i <= s_c & tos(data_width-1 DOWNTO 1);  -- Rotation right through carry. Watch out range limits.
                           carry_out <= tos(0);
           WHEN op_ROL  => alu_i <= tos(data_width-2 DOWNTO 0) & s_c;
                           carry_out <= tos(data_width-1);
           WHEN op_ZEQU => IF  tos=slice('0', data_width)  THEN        -- Equal zero?: State of tos.
                              alu_i <= (OTHERS => '1');
                           ELSE
                              alu_i <= (OTHERS => '0');
                           END IF;
           WHEN op_CC   => carry_out <= NOT s_c;
                           alu_i <= tos;
           WHEN OTHERS  => NULL;
           END CASE;
       WHEN op_NONE =>
           adder_mux <= '0' & nos;
           IF  with_tor='1'  THEN
              CASE  i_group  IS
              WHEN op_MULTS => adder_in1 <= '0' & tos(tos'high-1 DOWNTO 0) & '0';
                               IF  step_carry='1' AND tor_reg(tor_reg'high)='1'  THEN
                                  tor_new <= (tor_reg(tor_reg'high-1 DOWNTO 0) & tos(tos'high))+1;
                               ELSE
                                  tor_new <= tor_reg(tor_reg'high-1 DOWNTO 0) & tos(tos'high);
                               END IF;
              WHEN op_0DIVS => ovl_new <= '0';
                               carry_out <= '0';
                               alu_i <= nos;
                               tor_new <= stack;
              WHEN op_UDIVS => adder_in1 <= (s_c OR s_ovl) & tos;
                               adder_mux <= NOT('0' & nos);
                               carry_in <= '1';
                               tor_new <= tor_reg(data_width-2 DOWNTO 0) & carry;
                               IF  carry='1'  THEN
                                  alu_i <= sum(data_width-2 DOWNTO 0) & tor_reg(tor_reg'high);
                                  carry_out <= sum(data_width-1);
                               ELSE
                                  alu_i <= tos(data_width-2 DOWNTO 0) & tor_reg(tor_reg'high);
                                  carry_out <= tos(data_width-1);
                               END IF;
                               ovl_new <= NOT(step_carry XOR carry) OR s_ovl;
              WHEN op_LDIVS => adder_in1 <= (s_c OR s_ovl) & tos;
                               adder_mux <= NOT('0' & nos);
                               carry_in <= '1';
                               tor_new <= tor_reg(data_width-2 DOWNTO 0) & carry;
                               IF  carry='0'  THEN  alu_i <= tos;  END IF;
                               ovl_new <= tor_reg(data_width-1) OR s_ovl;
              WHEN op_SWAP  => alu_i <= nos;
              WHEN OTHERS => NULL;
              END CASE;
           END IF;
       WHEN OTHERS  =>
           adder_mux <= nos(nos'high) & nos; -- sign extend NOS
           CASE  i_group  IS
           WHEN op_ADD => carry_out <= carry;
                          ovl_new <= overflow;
           WHEN op_ADC => carry_in  <= s_c;
                          carry_out <= carry;
                          ovl_new <= overflow;
           WHEN op_SUB => adder_in1 <= NOT (tos(tos'high) & tos);
                          carry_in  <= '1';
                          carry_out <= carry;
                          ovl_new <= overflow;
           WHEN op_SSUB =>
                          adder_mux <= NOT (nos(nos'high) & nos);
                          carry_in  <= '1';
                          carry_out <= carry;
                          ovl_new <= overflow;
           WHEN op_AND => alu_i <= tos AND nos;
           WHEN op_OR  => alu_i <= tos OR  nos;
           WHEN op_XOR => alu_i <= tos XOR nos;
           WHEN op_NOS => alu_i <= nos;
           WHEN OTHERS => NULL;
           END CASE;
       END CASE;
   WHEN op_MEM =>
       CASE i_stack IS
       WHEN op_NONE =>
           CASE  i_group  IS
           WHEN op_RSTACK => IF  with_tor='1'     THEN  tor_new <= tos;    END IF;
           WHEN op_TOR    => IF  with_tor='1'     THEN  tor_new <= tos;    END IF;
           WHEN op_LOCAL  => IF  with_locals='1'  THEN  adder_mux <= '0' & rsp;  END IF;
           WHEN op_TASK   => IF  with_tasks='1'   THEN  adder_mux <= '0' & task; END IF;
           WHEN OTHERS    => NULL;
           END CASE;
       WHEN op_POP  => adder_mux <= slice(i_group(2), data_width-2) & i_group;
       WHEN op_PUSH => adder_mux <= slice(i_group(2), data_width-2) & i_group;
       WHEN op_BOTH =>
           CASE  i_group  IS
           WHEN op_RSTACK => IF  with_tor='1'     THEN  tor_new <= uBus_data;     END IF;
           WHEN op_LOCAL  => IF  with_locals='1'  THEN  adder_mux <= '0' & rsp;   END IF;
           WHEN op_TASK   => IF  with_tasks='1'   THEN  adder_mux <= '0' & task;  END IF;
           WHEN OTHERS    => NULL;
           END CASE;
       WHEN OTHERS => NULL;
       END CASE;
   WHEN OTHERS => -- op_USR
       IF  with_tor='1'  THEN
          tor_new <= pc; -- push_rstack;
       END IF;
   END CASE;
END IF;

IF  with_tor='0'  THEN  tor_new <= (OTHERS => '0');  END IF;

END PROCESS alu_output;

carry_status: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      s_c <= '0';
      s_ovl <= '0';
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        s_c <= carry_out;
        s_ovl <= ovl_new;
        IF  stBus.write='1'  THEN
           s_c <= stBus.tos(s_c_bit);
           s_ovl <= stBus.tos(s_ovl_bit);
        END IF;
     END IF;
   END IF;
END PROCESS carry_status;

stBus.flags(s_c_bit) <= s_c;
stBus.flags(s_ovl_bit) <= s_ovl;

------------------------------------------------------------------------------
-- TOR register (Top-Of-Returnstack
------------------------------------------------------------------------------

tor_register : PROCESS (clk, reset)
BEGIN
  IF  reset='1'  THEN
     tor_reg <= (OTHERS => '0');
  ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        tor_reg <= tor_new;
     END IF;
  END IF;
END PROCESS tor_register;

END rtl;
