-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore CLOCKS.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Phased clock generation for internal block_RAM control.
-- This file should be edited for technology specific additions
-- like e.g. the use of phase locked loop clock generators.
--
-- Each instruction execution cycle consists of 4 phases, which are derived
-- from SYSCLK. The derived clock signals are SYSCLK duty cycle dependet.
--
-- Phase 1: Tri-stating of uBus via the BUS_EN signal and instruction decode.
--
-- Phase 2: Data memory and return stack address computation. During memory reads,
--          RAM_CLK generates an active edges at the end of this phase in order to
--          latch the memory address for synchronous block-RAM.
--
-- Phase 3: Address computation for the next instruction. ROM_CLK generates an active
--          edge at the end of this phase in order to latch the memory address for
--          synchronous block-RAM.
--
-- Phase 4: Instruction fetch, saving computational results on the stack and writing
--          to memory with the rising edge of uCLK. On memory write cycles, RAM_CLK and
--          ROM_CLK generate their active edge at the end of this cycle.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     work.constants.ALL;

ENTITY clocks IS
PORT (reset     : IN    STD_LOGIC;
      sysclk    : IN    STD_LOGIC; -- at least twice the uClk frequency
      clk_en    : IN    STD_LOGIC;
      rom_write : IN    STD_LOGIC;
      rom_cache : IN    STD_LOGIC;
      ram_write : IN    STD_LOGIC;
      uClk      : OUT   STD_LOGIC;
      ramClk    : OUT   STD_LOGIC;
      ramWr     : OUT   STD_LOGIC;
      romClk    : OUT   STD_LOGIC;
      romWr     : OUT   STD_LOGIC;
      uBus      : INOUT data_port);
END clocks;

ARCHITECTURE rtl OF clocks IS

SIGNAL clock    : STD_LOGIC;
SIGNAL clk      : STD_LOGIC;
SIGNAL bus_res  : STD_LOGIC;
SIGNAL bus_en   : STD_LOGIC;
SIGNAL ram_res  : STD_LOGIC;
SIGNAL ram_clk  : STD_LOGIC;
SIGNAL ram_wr   : STD_LOGIC;
SIGNAL rom_res  : STD_LOGIC;
SIGNAL rom_clk  : STD_LOGIC;
SIGNAL rom_wr   : STD_LOGIC;

BEGIN

clock <= sysclk AFTER 2 ns; -- additional divider if needed to generate CLOCK from SYSCLK

-----------------------------------------------------------------------
-- uClk the predominant uCore clock
-----------------------------------------------------------------------

uClk_oscillator: PROCESS (reset, clock)
BEGIN
   IF  reset='1'  THEN
      clk <= '1';
   ELSIF  rising_edge(clock)  THEN
      clk <= NOT clk;
   END IF;
END PROCESS uClk_oscillator;

uClk <= clk;

-----------------------------------------------------------------------
-- synchronous logic to generate uBus.en
-- 'Z'-states uBus for 1/4 uClk cycle to prevent cross-currents
-- in this 3state buffer implementation.
-----------------------------------------------------------------------

bus_res <= reset OR (NOT bus_en AND NOT clock);

tristate_holdoff: PROCESS (bus_res, clock)
BEGIN
  IF  bus_res='1'  THEN
     bus_en <= '1';
  ELSIF  rising_edge(clock)  THEN
     IF  clk_en='1' THEN
        bus_en <= clk;
     END IF;
  END IF;
END PROCESS tristate_holdoff;

uBus.en <= bus_en;

-----------------------------------------------------------------------
-- ramClk, sel_ram gated, generating different edges for read and write
-----------------------------------------------------------------------

ram_res <= ram_clk AND clk AND NOT clock AND uBus.sel_dcache;

ramClk_oscillator: PROCESS (clock, ram_res)
BEGIN
   IF ram_res='1'  THEN
      ram_clk <= '0';
   ELSIF  rising_edge(clock)  THEN
      IF  clk_en='1'  THEN
         IF  ram_write='0'  THEN
            ram_clk <= '1';
         ELSIF  clk='0'  THEN
            ram_clk <= '1';
         END IF;
      IF  reset='1'  THEN
         ram_clk <= '1';
      END IF;
      END IF;
   END IF;
END PROCESS ramClk_oscillator;

ramClk <= ram_clk;

-----------------------------------------------------------------------
-- ramWr, sel_ram and sel_ext gated, asynchronous write clock
-- for external RAM
-----------------------------------------------------------------------

ram_write_clk : PROCESS (reset, clock)
BEGIN
  IF  reset='1'  THEN
     ram_wr <= '0';
  ELSIF  rising_edge(clock)  THEN
     IF  clk_en='1'  THEN
        ram_wr <= clk AND ram_write AND (uBus.sel_ram OR uBus.sel_ext);
     END IF;
  END IF;
END PROCESS ram_write_clk;

ramWr <= ram_wr;

-----------------------------------------------------------------------
-- romClk, rom_cache gated, generating different edges for read and write
-----------------------------------------------------------------------

rom_res <= reset OR (NOT rom_clk AND NOT rom_write AND NOT clock);

romClk_oscillator: PROCESS (clock, rom_res)
BEGIN
   IF  rom_res='1'  THEN
      rom_clk <= '1';
   ELSIF  rising_edge(clock)  THEN
      IF  rom_cache='1' AND clk_en='1'  THEN
         rom_clk <= NOT clk;
      END IF;
   END IF;
END PROCESS romClk_oscillator;

romClk <= rom_clk;

-----------------------------------------------------------------------
-- romWr, (NOT rom_cache) gated, asynchronous write clock for external RAM
-----------------------------------------------------------------------

rom_write_clk : PROCESS (reset, clock)
BEGIN
  IF  reset='1'  THEN
     rom_wr <= '0';
  ELSIF  rising_edge(clock)  THEN
     IF  clk_en='1'  THEN
        rom_wr <= clk AND rom_write AND NOT rom_cache;
     END IF;
  END IF;
END PROCESS rom_write_clk;

romWr <= rom_wr;

END rtl;