-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore DEBUG_CENTRONICS.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- Umbilical, asynchronous debug interface with full duplex handshake.
-- It consists of a program memory loader and an IN and OUT register,
-- which can be written and read by the host asynchronously and read
-- and written by the core thru DEBUG_REGISTER.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY debugger IS
PORT (done       : IN    STD_LOGIC;
      clk        : IN    STD_LOGIC;
      clk_en     : IN    STD_LOGIC;
      uBus       : INOUT data_port;
      sel_data   : OUT   STD_LOGIC;
      debug_rom  : INOUT program_port;
      umbilical  : INOUT debug_port;
      in_full    : OUT   STD_LOGIC;
      out_empty  : OUT   STD_LOGIC;
      soft_reset : OUT   STD_LOGIC;
      out_except : OUT   STD_LOGIC;
      rom_load   : OUT   STD_LOGIC);
END debugger;

ARCHITECTURE rtl OF debugger IS

ATTRIBUTE syn_keep : BOOLEAN;

SIGNAL init       : STD_LOGIC;    -- host initiates program loading
SIGNAL in_str     : STD_LOGIC;    -- host signals valid data
SIGNAL in_ack     : STD_LOGIC;    -- target captures data on rising_edge
SIGNAL in_data    : STD_LOGIC_VECTOR(7 DOWNTO 0); -- data from host
SIGNAL in_rd      : STD_LOGIC; ATTRIBUTE syn_keep OF in_rd : SIGNAL IS true;   -- target reads IN_REG
SIGNAL in_reg     : data_bus;
SIGNAL in_ctr     : STD_LOGIC_VECTOR(log2(in_nibbles)-1 DOWNTO 0); -- #nibbles to transfer from host
SIGNAL in_prev    : STD_LOGIC;
SIGNAL in_reset   : STD_LOGIC;
SIGNAL cold       : STD_LOGIC;

SIGNAL addr_ctr   : prog_addr;    -- incremental debugROM address for loading
SIGNAL write      : STD_LOGIC;    -- write signal for debugROM

SIGNAL out_str    : STD_LOGIC;    -- target signals valid data
SIGNAL out_ack    : STD_LOGIC;    -- host captures data on rising_edge
SIGNAL out_reg    : data_bus;     -- out_data
SIGNAL out_wr     : STD_LOGIC;    -- target writes OUT_REG
SIGNAL out_ctr    : STD_LOGIC_VECTOR(log2(data_width)-1 DOWNTO 0); -- #bits to transfer to host

TYPE in_states IS (idle, loading, receiving, full);
SIGNAL in_state  : in_states;

TYPE out_states IS (idle, sending);
SIGNAL out_state : out_states;
 
BEGIN

in_str  <= umbilical.in_str;   -- for the synthesizer
in_data <= umbilical.in_data;  -- for the synthesizer
init    <= umbilical.init;     -- for the synthesizer

coldboot : PROCESS (done, clk)
BEGIN
  IF  done='0'  THEN
     cold <= '1';
  ELSIF  rising_edge(clk)  THEN
     IF  in_state=loading  THEN
        cold <= '0';
     END IF;
  END IF;   
END PROCESS coldboot;   

-----------------------------------------------------------------------
-- loading the program memory
-- and listening to the host  
-----------------------------------------------------------------------

in_rd <= NOT uBus.write WHEN  uBus.internal='1' AND uBus.addr=DEBUG_REGISTER  ELSE '0';

uBus.data <= in_reg WHEN  uBus.en='1' AND in_rd='1'  ELSE (OTHERS => 'Z') AFTER 5 ns;
sel_data <= in_rd;

loadROM: PROCESS(clk, done)
BEGIN
  IF  done='0'  THEN
     in_state <= idle;
     addr_ctr <= (OTHERS => '0');
       in_reg <= (OTHERS => '0');
       in_ack <= '0';
      in_prev <= '0';
   soft_reset <= '0';
        write <= '0';
       in_ctr <= (OTHERS => '0');
  ELSIF  rising_edge(clk)  THEN
     write <= '0';
     soft_reset <= '0';
     in_ack <= in_str;
     IF  init='1'  THEN
        IF  in_str='1' AND in_ack='0'   THEN
           in_prev <= in_data(0);
        END IF;
        IF  in_str='0' AND in_ack='1'  THEN
           IF  in_prev/=in_data(0)  THEN
              addr_ctr <= (OTHERS => '0');
              in_state <= loading;
              soft_reset <= in_prev;
           END IF;
        END IF;
     END IF;
     CASE in_state IS
     WHEN loading => 
         IF  in_str='1' AND in_ack='0'  THEN
            IF  write='0'  THEN
               write <= '1';
               in_ack <= '0';
            ELSE
               addr_ctr <= addr_ctr+1;
            END IF;
         END IF;
         IF  in_str='0' AND init='0'  THEN
            in_state <= idle;
         END IF;
     WHEN receiving =>
         IF  in_ctr=0  THEN
            in_state <= full;
         END IF;
         IF  (in_str AND NOT in_ack)='1'  THEN
            in_reg <= in_reg(data_width-9 DOWNTO 0) & in_data;
            in_ctr <= in_ctr-1;
         END IF;
     WHEN full => 
         IF  in_rd='1' AND clk_en='1'  THEN
           in_state <= idle;
         END IF;
         IF  in_ack='0'  THEN
            in_ack <= '0';
         END IF; 
     WHEN idle =>
         IF  init='0' AND in_str='1' AND in_ack='0'  THEN
           in_reg <= in_reg(data_width-9 DOWNTO 0) & in_data;
           in_ctr <= to_vec(in_nibbles-1, log2(in_nibbles));
           in_state <= receiving;
         END IF;
     WHEN OTHERS => in_state <= idle;
     END CASE;
   END IF;
END PROCESS loadROM;

debug_rom.write <= write;
debug_rom.dout  <= in_data;
debug_rom.addr  <= addr_ctr;

umbilical.in_ack <= in_ack;
umbilical.busy   <= '1' WHEN  in_state=loading OR in_state=full  ELSE '0';
rom_load     <= '1' WHEN  in_state=loading OR cold='1'       ELSE '0';
in_full      <= '1' WHEN  in_state=full                      ELSE '0';

-----------------------------------------------------------------------
-- talking to the host 
-----------------------------------------------------------------------

out_ack <= umbilical.out_ack;     -- for the synthesizer

out_wr  <= uBus.write WHEN  uBus.internal='1' AND uBus.addr=DEBUG_REGISTER  ELSE '0';

debug_out: PROCESS (clk, done)
BEGIN
  IF  done='0'  THEN
       out_str <= '0';
       out_reg <= (OTHERS => '0');
       out_ctr <= (OTHERS => '1');
     out_state <= idle;
  ELSIF  rising_edge(clk)  THEN
    out_str <= NOT out_ack;
    CASE out_state IS
    WHEN idle =>
        out_str <= '0';
        IF  out_wr='1' AND clk_en='1'  THEN   -- send a new word
            out_reg <= uBus.data;
            out_ctr <= to_vec(data_width-1, log2(data_width));
          out_state <= sending;
        END IF;
    WHEN sending =>
        IF  (out_ack AND out_str)='1'  THEN  -- shift next bit on out_ack
          IF  out_ctr=0  THEN  
             out_state <= idle;
          ELSE  
             out_reg <= out_reg(out_reg'high-1 DOWNTO 0) & '0';
             out_ctr <= out_ctr-1;
          END IF;
        END IF;
    WHEN OTHERS => out_state <= idle;
    END CASE;
  END IF;
END PROCESS debug_out;

umbilical.out_str  <= out_str;
umbilical.out_data <= out_reg(out_reg'high);

out_empty <= '1' WHEN  out_state=idle  ELSE '0';

out_except <= '1' WHEN  (out_wr='1' AND out_state=sending)
                     OR (in_rd='1' AND in_state/=full)  ELSE
              '0';

END rtl;
