-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore RSTACK.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- The uCore return stack. It uses the topmost data
-- ram area and and can be addressed using absolute or return
-- stack pointer relative addresses when with_locals is true.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY r_stack IS
PORT (reset     : IN    STD_LOGIC; 
      clk       : IN    STD_LOGIC;
      clk_en    : IN    STD_LOGIC;
      inst      : IN    inst_bus;
      stBus     : IN    status_rec;
      condition : IN    STD_LOGIC;
      token     : IN    STD_LOGIC;
      tor_zero  : IN    STD_LOGIC;
      uBus      : INOUT data_port;
      alu       : IN    data_bus;
      ip        : IN    data_bus;
      rsp       : OUT   data_bus;
      l_addr    : OUT   data_bus);
END r_stack;

ARCHITECTURE rtl OF r_stack IS

SIGNAL i_lit      : STD_LOGIC;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;

SIGNAL rsp_reg    : rs_addr;
SIGNAL rsp_new    : rs_addr;
SIGNAL rsp_minus  : rs_addr;
SIGNAL rsp_plus   : rs_addr;
SIGNAL addr       : data_addr;
SIGNAL uBus_data  : data_bus;

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);

rsp <= '0' & slice('1', data_width-(rs_addr_width+tasks_addr_width)-1) & rsp_reg;
l_addr <= '0' & slice('1', data_width-(rs_addr_width+tasks_addr_width)-1) & rsp_minus  WHEN  with_tor='1'  ELSE
          '0' & slice('1', data_width-(rs_addr_width+tasks_addr_width)-1) & rsp_reg;
uBus.addr <= addr;
uBus_data <= uBus.data;

rsp_moves : PROCESS (rsp_reg)
BEGIN
  IF  tasks_addr_width=0  THEN
    rsp_minus <= rsp_reg-1;
    rsp_plus <= rsp_reg+1;
  ELSE
    rsp_minus <= rsp_reg(tasks_addr_width+rs_addr_width-1 DOWNTO rs_addr_width) & rsp_reg(rs_addr_width-1 DOWNTO 0)-1;
    rsp_plus <= rsp_reg(tasks_addr_width+rs_addr_width-1 DOWNTO rs_addr_width) & rsp_reg(rs_addr_width-1 DOWNTO 0)+1;
  END IF;
END PROCESS rsp_moves;

------------------------------------------------------------------------------
-- return stack grows from higher to lower addresses
------------------------------------------------------------------------------

rstack_control: PROCESS (rsp_reg, i_lit, i_type, i_stack, i_group, condition, uBus_data, alu, rsp_minus, rsp_plus, tor_zero, ip, token)

PROCEDURE push_rstack IS
BEGIN
  rsp_new <= rsp_minus;
  addr <= slice('1', data_addr_width-(rs_addr_width+tasks_addr_width)) & rsp_minus;
END push_rstack;

PROCEDURE pop_rstack IS
BEGIN
  rsp_new  <= rsp_plus;
  addr <= slice('1', data_addr_width-(rs_addr_width+tasks_addr_width)) & rsp_reg;
END pop_rstack;

BEGIN

addr <= alu(data_addr_width-1 DOWNTO 0);
rsp_new <= rsp_reg;

IF  i_lit='0'  THEN
   CASE  i_type  IS
   WHEN op_BRA =>
       CASE i_stack IS
       WHEN op_NONE => IF  condition='1'  THEN
                          pop_rstack;
                       END IF;
       WHEN op_POP  => IF  with_tor='1' AND i_group=op_NZERO AND tor_zero='1'  THEN -- end of NEXT instruction
                          pop_rstack;
                       END IF;
       WHEN op_PUSH =>
          CASE  i_group  IS
          WHEN op_EXC    => push_rstack;
          WHEN op_INT    => push_rstack;
          WHEN op_IRET   => pop_rstack;
          WHEN op_QOVL   => IF  stBus.flags(s_ovl_bit)='1'  THEN
                               push_rstack;
                            END IF;
          WHEN op_THREAD => IF  with_ip='1'  THEN
                               addr <= ip(data_addr_width-1 DOWNTO 0);
                            END IF;
          WHEN op_TOKEN  => IF  with_tokens='1' AND with_ip='1'  THEN
                               IF  token='0'  THEN
                                  push_rstack;
                               ELSE
                                  addr <= ip(data_addr_width-1 DOWNTO 0);
                               END IF;
                            END IF;
          WHEN OTHERS => NULL;
          END CASE;
       WHEN op_BOTH => IF  condition='1'  THEN
                          push_rstack;
                       END IF;
       WHEN OTHERS => NULL;
       END CASE;
   WHEN op_ALU => IF  with_tor='1' AND i_stack=op_NONE AND i_group=op_0DIVS  THEN
                     push_rstack;
                  END IF;
   WHEN op_MEM =>
       CASE i_stack IS
       WHEN op_NONE =>
          CASE  i_group  IS
          WHEN op_RSTACK => push_rstack;
          WHEN op_TOR    => addr <= slice('1', data_addr_width-(rs_addr_width+tasks_addr_width)) & rsp_reg; -- r!
          WHEN op_RSP    => rsp_new <= uBus_data(tasks_addr_width+rs_addr_width-1 DOWNTO 0);
          WHEN OTHERS    => NULL;
          END CASE;
       WHEN op_POP  => NULL;
       WHEN op_PUSH => NULL;
       WHEN OTHERS  => -- op_BOTH
           CASE  i_group  IS
           WHEN op_RSTACK => pop_rstack;
           WHEN op_TOR    => IF  with_tor='0'  THEN
                                addr <= slice('1', data_addr_width-(rs_addr_width+tasks_addr_width)) & rsp_reg; -- r@
                             END IF;
           WHEN OTHERS    => NULL;
           END CASE;
       END CASE;
   WHEN OTHERS => -- op_USR
      push_rstack;
   END CASE;
END IF;
END PROCESS rstack_control;

------------------------------------------------------------------------------
-- return stack grows from higher to lower addresses
------------------------------------------------------------------------------

rsp_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      rsp_reg <= (OTHERS => '1');
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        rsp_reg <= rsp_new;
     END IF;
   END IF;
END PROCESS rsp_register;

END rtl;
