-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore SEQUENCER.VHD 1.20
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
--
-- The sequencer generates the address for the next
-- instruction fetch from program memory. It realises
-- interrupt processing and the exception mechanism for
-- efficient (transputer style) multitasking.

LIBRARY IEEE;
USE     IEEE.STD_LOGIC_1164.ALL;
USE     IEEE.STD_LOGIC_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY sequencer IS
PORT (reset     : IN    STD_LOGIC; 
      clk       : IN    STD_LOGIC;
      clk_en    : IN    STD_LOGIC;
      except    : IN    STD_LOGIC;
      int       : IN    STD_LOGIC;
      stBus     : IN    status_rec;
      uBus      : IN    data_port;
      alu       : IN    data_bus;
      tor       : IN    data_bus;
      tos       : IN    data_bus;
      instrom   : IN    inst_bus;
      instaddr  : OUT   prog_addr;
      inst      : OUT   inst_bus;
      condition : OUT   STD_LOGIC;
      pc        : OUT   data_bus;
      token     : OUT   STD_LOGIC;
      ip        : OUT   data_bus;
      ip_plus   : OUT   data_bus);
END sequencer;

ARCHITECTURE rtl OF sequencer IS

SIGNAL inst_reg   : inst_bus;
SIGNAL i_lit      : STD_LOGIC;
SIGNAL s_lit      : STD_LOGIC;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;
SIGNAL pc_reg     : prog_addr;
SIGNAL paddr      : prog_addr;
SIGNAL cond_i     : STD_LOGIC;
SIGNAL uBus_data  : data_bus;
SIGNAL ip_reg     : data_addr;
SIGNAL flag_next  : STD_LOGIC;
SIGNAL tor_zero   : STD_LOGIC;
SIGNAL token_i    : STD_LOGIC;   
SIGNAL s_ovl      : STD_LOGIC;

BEGIN

inst      <= inst_reg;
instaddr  <= paddr;
condition <= cond_i;
i_lit     <= inst_reg(inst_width-1);
s_lit     <= stBus.flags(s_lit_bit);
i_type    <= inst_reg(inst_width-2 DOWNTO inst_width-3);
i_stack   <= inst_reg(inst_width-4 DOWNTO inst_width-5);
i_group   <= inst_reg(inst_width-6 DOWNTO 0);
s_ovl     <= stBus.flags(s_ovl_bit);
pc        <= slice('0', data_width-prog_addr_width) & pc_reg;
ip        <= slice('0', data_width-data_addr_width) & ip_reg   WHEN  with_ip='1'  ELSE  (OTHERS => '0');
ip_plus   <= slice('0', data_width-data_addr_width) & ip_reg+1 WHEN  with_ip='1'  ELSE  (OTHERS => '0');
token     <= token_i;

uBus_data <= uBus.data;

------------------------------------------------------------------------------
-- sequencer_control: Generates control lines for the sequencer entity
-- The opcodes' fields are defined in instruction register as follows:
--   7   6   5   4   3   2   1   0
-- |---|---|---|---|---|---|---|---|
-- |   |   |   |   |   |   |   |   |
-- |---|---|---|---|---|---|---|---|
-- |LIT|  TYPE | STACK |    GROUP  |
-- |---|-------|-------|-----------|
--
-- In the process the op-field are preceded by "i_" .
-- The different "op_XXX" are defined in constants.vhdl.
--
------------------------------------------------------------------------------

tor_zero <= '1' WHEN  with_tor='1' AND tor=0  ELSE '0';

flag_next <= tor_zero WHEN  with_tor='1' AND inst_reg=NEXT_OP  ELSE stBus.flags(s_z_bit);

WITH  i_group  SELECT
cond_i <=  stBus.flags(s_z_bit)    WHEN op_ZERO,
       NOT flag_next               WHEN op_NZERO,
           stBus.flags(s_n_bit)    WHEN op_SIGN,
       NOT stBus.flags(s_n_bit)    WHEN op_NSIGN,
       NOT stBus.flags(s_ovl_bit)  WHEN op_NOVL,
       NOT stBus.flags(s_c_bit)    WHEN op_NCARRY,
                              '0'  WHEN op_NEVER,
                              '1'  WHEN OTHERS;  --op_ALWAYS,

------------------------------------------------------------------------------
-- Build program's memory address. 
-- You have the following choices:
-- rom address = pc_reg
-- rom address = uBus_data
-- rom address = 0x0004 address of interrupt service routine
-- rom address = 0x0008 address of task service routine
-- and more...
------------------------------------------------------------------------------

addr_select: PROCESS (inst_reg, i_lit, s_lit, i_type, i_stack, i_group, cond_i, pc_reg, uBus_data, alu, tor, s_ovl)
BEGIN

paddr <= pc_reg;

IF  i_lit='0'  THEN
   CASE  i_type  IS
   WHEN op_BRA =>
      IF  cond_i='1'  THEN
         IF  with_tor='1'  THEN
            paddr <=       tor(prog_addr_width-1 DOWNTO 0);
         ELSE
            paddr <= uBus_data(prog_addr_width-1 DOWNTO 0);
         END IF;
         CASE i_stack IS
         WHEN op_POP  =>
            IF  s_lit='1'  THEN
               paddr <= alu(prog_addr_width-1 DOWNTO 0);
            ELSE
               paddr <= uBus_data(prog_addr_width-1 DOWNTO 0);
            END IF;
         WHEN op_BOTH => paddr <= alu(prog_addr_width-1 DOWNTO 0);
         WHEN OTHERS  => NULL;
         END CASE;
      END IF;
      IF  i_stack=op_PUSH  THEN
         paddr <= pc_reg;
         CASE i_group IS
         WHEN op_EXC  => paddr <= addr_esr;
         WHEN op_INT  => paddr <= addr_isr;
         WHEN op_QOVL => IF  s_ovl='1'  THEN
                            paddr <= addr_ovl;
                         END IF;
         WHEN op_IRET => IF  with_tor='1'  THEN
                            paddr <=       tor(prog_addr_width-1 DOWNTO 0);
                         ELSE
                            paddr <= uBus_data(prog_addr_width-1 DOWNTO 0);
                         END IF;
         WHEN op_THREAD | op_TOKEN =>
                         IF  with_ip='1' AND uBus_data(data_width-1)='1'  THEN
                            paddr <= uBus_data(prog_addr_width-1 DOWNTO 0);
                         END IF;
         WHEN OTHERS  => NULL;
         END CASE;
      END IF;
   WHEN op_USR =>
      paddr <= slice('0', (prog_addr_width-(inst_width-3+usr_vect_width))) & inst_reg(inst_width-4 DOWNTO 0) & slice('0', usr_vect_width);
   WHEN OTHERS => NULL;
   END CASE;
END IF;

END PROCESS addr_select;

-----------------------------------------------------------------------
-- pc_register (program counter)
-- (a) reset = 1 : pc_reg = 0 
-- (b) clk valid : pc_reg = addr+1
-- (c) clk valid & int = 1: pc_reg = addr (no increment during an interrupt acknowlege)
------------------------------------------------------------------------
pc_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      pc_reg <= addr_reset;
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
       pc_reg <= paddr+1;
       IF  inst_reg=THREAD_OP AND uBus_data(data_width-1)='0'  THEN
          pc_reg <= paddr-1;
       END IF;
       IF  int='1'  THEN
          pc_reg <= paddr;
       END IF;
     END IF;
   END IF;
END PROCESS pc_register;

-----------------------------------------------------------------------
-- inst_register (instruction register)
-- (a) reset = 1:  inst = NOP 
-- (b) clk valid:  inst = instrom
-- (c) clk valid & int  = 1: inst = INT_OP  (BRA BOTH INT)
-- (d) clk valid & trap = 1: inst = EXC_OP  (BRA BOTH EXC)
-----------------------------------------------------------------------
inst_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      inst_reg <= NO_OP;
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1' OR except='1'  THEN
       inst_reg <= instrom;
       IF  int='1'  THEN
          inst_reg <= INT_OP;
       END IF;
       IF  except='1'  THEN
          inst_reg <= EXC_OP;
       END IF;
     END IF;
   END IF;
END PROCESS inst_register;

-----------------------------------------------------------------------
-- ip_register (interpreter pointer register)
-- works with instructions "(ip+) ip+'"
-- as an innter interpreter for tag threaded code
-----------------------------------------------------------------------
make_ip_reg: IF  with_ip='1'  GENERATE
   ip_register : PROCESS (clk, reset)
   BEGIN
     IF  reset='1'  THEN
        ip_reg <= (OTHERS => '0');
     ELSIF  rising_edge(clk)  THEN
        IF  clk_en='1'  THEN
          IF  inst_reg=THREAD_OP  THEN
             IF  uBus_data(data_width-1)='1'  THEN
                ip_reg <= ip_reg+1;                                 -- its a call to code
             ELSE
                ip_reg <= uBus_data(data_addr_width-1 DOWNTO 0);    -- its a call to threaded code
             END IF;
          END IF;
          IF  with_tokens='1' AND inst_reg=TOKEN_OP AND token_i='1'  THEN
             ip_reg <= tos(data_addr_width-1 DOWNTO 0);             -- call to token
          END IF;
          IF  inst_reg=IPSTORE_OP  THEN
             ip_reg <= uBus_data(data_addr_width-1 DOWNTO 0);
          END IF;
        END IF;
     END IF;
   END PROCESS ip_register;
END GENERATE make_ip_reg;

token_i <= '1' WHEN  with_tokens='1' AND with_ip='1' AND ip_reg(data_addr_width-1 DOWNTO token_width)=0  ELSE  '0';

END rtl;
